@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2016 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2017 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 
 # This file is part of qutebrowser.
 #
@@ -23,7 +23,6 @@ import sys
 import json
 
 import qutebrowser
-from qutebrowser.utils import log
 try:
     from qutebrowser.misc.checkpyver import check_python_version
 except ImportError:
@@ -38,8 +37,9 @@ except ImportError:
         sys.stderr.flush()
         sys.exit(100)
 check_python_version()
+from qutebrowser.utils import log
 
-import argparse
+import argparse  # pylint: disable=wrong-import-order
 from qutebrowser.misc import earlyinit
 
 
@@ -64,8 +64,7 @@ def get_argparser():
                         help="How URLs should be opened if there is already a "
                              "qutebrowser instance running.")
     parser.add_argument('--backend', choices=['webkit', 'webengine'],
-                        help="Which backend to use (webengine backend is "
-                             "EXPERIMENTAL!).")
+                        help="Which backend to use.")
     parser.add_argument('--enable-webengine-inspector', action='store_true',
                         help="Enable the web inspector for QtWebEngine. Note "
                         "that this is a SECURITY RISK and you should not "
@@ -103,10 +102,6 @@ def get_argparser():
                        help="Silently remove unknown config options.")
     debug.add_argument('--nowindow', action='store_true', help="Don't show "
                        "the main window.")
-    debug.add_argument('--debug-exit', help="Turn on debugging of late exit.",
-                       action='store_true')
-    debug.add_argument('--pdb-postmortem', action='store_true',
-                       help="Drop into pdb on exceptions.")
     debug.add_argument('--temp-basedir', action='store_true', help="Use a "
                        "temporary basedir.")
     debug.add_argument('--no-err-windows', action='store_true', help="Don't "
@@ -118,6 +113,9 @@ def get_argparser():
                        action='append')
     debug.add_argument('--qt-flag', help="Pass an argument to Qt as flag.",
                        nargs=1, action='append')
+    debug.add_argument('--debug-flag', type=debug_flag_error, default=[],
+                       help="Pass name of debugging feature to be turned on.",
+                       action='append', dest='debug_flags')
     parser.add_argument('command', nargs='*', help="Commands to execute on "
                         "startup.", metavar=':command')
     # URLs will actually be in command
@@ -131,7 +129,7 @@ def directory(arg):
         raise argparse.ArgumentTypeError("Invalid empty value")
 
 
-def logfilter_error(logfilter: str):
+def logfilter_error(logfilter):
     """Validate logger names passed to --logfilter.
 
     Args:
@@ -145,15 +143,25 @@ def logfilter_error(logfilter: str):
                 logfilter, ', '.join(log.LOGGER_NAMES)))
 
 
+def debug_flag_error(flag):
+    """Validate flags passed to --debug-flag.
+
+    Available flags:
+        debug-exit: Turn on debugging of late exit.
+        pdb-postmortem: Drop into pdb on exceptions.
+    """
+    valid_flags = ['debug-exit', 'pdb-postmortem']
+
+    if flag in valid_flags:
+        return flag
+    else:
+        raise argparse.ArgumentTypeError("Invalid debug flag - valid flags: {}"
+                                         .format(', '.join(valid_flags)))
+
+
 def main():
     parser = get_argparser()
-    if sys.platform == 'darwin' and getattr(sys, 'frozen', False):
-        # Ignore Mac OS X' idiotic -psn_* argument...
-        # http://stackoverflow.com/questions/19661298/
-        # http://sourceforge.net/p/cx-freeze/mailman/message/31041783/
-        argv = [arg for arg in sys.argv[1:] if not arg.startswith('-psn_0_')]
-    else:
-        argv = sys.argv[1:]
+    argv = sys.argv[1:]
     args = parser.parse_args(argv)
     if args.json_args is not None:
         # Restoring after a restart.
